import json
import traceback

from dataclasses import asdict

from ._models import TestLiteTestReport, FixtureRunResult, TestLiteFixtureReport


class MakeDict:

    def __init__(self):
        self.remake_object = None


    def _make_dict_from_FixtureRunResult(self, obj):
        if isinstance(obj, FixtureRunResult):
            item = asdict(obj)
            item.update({
                'error': "".join(traceback.format_exception_only(type(obj.error), obj.error)).strip() if obj.error is not None else None
            })
            return item
        else:
            raise Exception('Its not a FixtureRunResult clas')
    

    def remake(self, obj, key):
        if isinstance(obj, TestLiteFixtureReport):
            item = {}
            if key == 'before':
                item.update({
                    'name': obj.name,
                    'start_time': obj.before_start_time,
                    'stop_time': obj.before_stop_time,
                    'duration': obj.before_duration,
                    'status': self._make_dict_from_FixtureRunResult(obj.before_status)
                })
            if key == 'after':
                item.update({
                    'name': obj.name,
                    'start_time': obj.after_start_time,
                    'stop_time': obj.after_stop_time,
                    'duration': obj.after_duration,
                    'status': self._make_dict_from_FixtureRunResult(obj.after_status)
                })
            return item
        else:
            raise Exception('Its not TestLiteFixtureReport class')
        

    def make_serializable_dict_from_fixtures_dict(self, fixtures_dict):
        self.remake_object = fixtures_dict
        if isinstance(self.remake_object, dict):
            for key, value in self.remake_object.items():
                for i, item in enumerate(value):
                    self.remake_object[key][i] = self.remake(item, key)
            return self.remake_object
        else:
            raise Exception('Its not a dict')



class TestReportJSONEncoder(json.JSONEncoder):

    def default(self, o):
        if isinstance(o, TestLiteTestReport):
            item = asdict(o)
            del item['_fixturelist']
            item.update({
                'parametrize_name': str(o.parametrize_name),
                'startime_readable': str(o.startime_readable),
                'stoptime_readable': str(o.stoptime_readable), 
                'duration': float(o.duration),
                'duration': float(o.duration),
                'fixtures': MakeDict().make_serializable_dict_from_fixtures_dict(o.fixtures)
            })
            return item
        return super().default(o)


import pytest
from testlite._reports import (
    TestLiteTestReport,
    TestLiteTestReports,
    TestLiteFixtureReport,
    TestLiteFixtureReports,
    TestLiteReportManager
)
from testlite._helper import get_time
from testlite._reports import fixture_after_save
from testlite._models import STATUS
from testlite._Testlite import TestLite_testcase_key, get_step_number_with_error


def pytest_addoption(parser):
    group = parser.getgroup('TestLitePytest')
    group.addoption(
        '--testsuite',
        action='store',
        dest='testsuite',
        help='Set TestSuite Key. If the option is specified, then at the end we will try to send the results to TestLite'
    )
    group.addoption(
        '--save_json',
        action='store',
        dest='savejson',
        default=None,
        help='After end of testing saving JSON file with report'
    )

    parser.addini('HELLO', 'Dummy pytest.ini setting')



def pytest_configure(config):
    config.TSTestReports = TestLiteTestReports



@pytest.hookimpl(hookwrapper=True)
def pytest_runtest_makereport(item, call):
    outcome = yield
    report: pytest.TestReport = outcome.get_result()
    test_report: TestLiteTestReport = item.config.TSTestReports.get_test_report(report.nodeid)
    test_report.testcase_key = TestLite_testcase_key(item)


    if report.when == 'setup':

        test_report.params = item.callspec.params if hasattr(item, 'callspec') else None

        if report.skipped == True:
            test_report.status = STATUS.SKIP
            test_report.skipreason = report.longrepr[2]

            test_report.startime_timestamp = report.start
            test_report.stoptime_timestamp = report.stop


        if report.failed == True:
            test_report.precondition_status = STATUS.ERROR
            test_report.status = STATUS.ERROR
            test_report.report = report.longreprtext

            test_report.startime_timestamp = report.start
            test_report.stoptime_timestamp = report.stop


        if report.passed == True:
            test_report.precondition_status = STATUS.PASSED

            test_report.startime_timestamp = report.start


    if test_report.status != STATUS.SKIP and report.when == 'call':
        if report.passed == True:
            test_report.status = STATUS.PASSED


        if report.failed == True:
            test_report.step_number_with_error = get_step_number_with_error(report.longreprtext)
            test_report.status = STATUS.FAIL
            test_report.report = report.longreprtext


    if test_report.status != STATUS.SKIP and report.when == 'teardown':
        if report.failed == True:
            test_report.postcondition_status = STATUS.ERROR
            test_report.report = report.longreprtext
        if report.passed == True:
            test_report.postcondition_status = STATUS.PASSED

        test_report.log = report.caplog
        test_report.stoptime_timestamp = report.stop

    item.config.TSTestReports.save_test_report(test_report) 



@pytest.hookimpl(hookwrapper=True)
def pytest_fixture_setup(fixturedef, request):
    if fixturedef.baseid != '':
        test_function_name = str(request._pyfuncitem).split(' ')[1][:-1]
        id = f'{test_function_name}::{request.fixturename}'
        nodeid = f'{fixturedef.baseid}::{test_function_name}'

        fixture_report = TestLiteFixtureReports.get_fixture_report(id, nodeid)
        fixture_report.before_start_time = get_time()
        TestLiteFixtureReports.save_fixture_report(id, fixture_report)

        outcome = yield

        finalizers = getattr(fixturedef, '_finalizers', [])
        for index, finalizer in enumerate(finalizers):
            finalizers[index] = fixture_after_save(finalizer, id, nodeid)
            
        fixture_report.before_stop_time = get_time()
        TestLiteFixtureReports.save_fixture_report(id, fixture_report)
    else:
        yield



def pytest_fixture_post_finalizer(fixturedef, request):
    if fixturedef.baseid != '':
        test_function_name = str(request._pyfuncitem).split(' ')[1][:-1]
        id = f'{test_function_name}::{request.fixturename}'
        nodeid = f'{fixturedef.baseid}::{test_function_name}'

        fixture_report = TestLiteFixtureReports.get_fixture_report(id, nodeid)
        fixture_report.name = request.fixturename
        fixture_report.cached_result = fixturedef.cached_result
        TestLiteFixtureReports.save_fixture_report(id, fixture_report)



def pytest_sessionfinish(session: pytest.Session, exitstatus):
    if hasattr(session.config, "workerinput"):
        TestLiteReportManager().save_report()
    else:
        TestLiteReportManager().save_report()
        final_report = TestLiteReportManager().get_reports()
        if session.config.getoption('--save_json') is not None:
            final_report.save_json_file(session.config.getoption("--save_json"))
        if session.config.getoption('--testsuite'):
            final_report.send_json_in_TestLite(testsuite=session.config.getoption('--testsuite'))



def test_key(TestLite_id):

    def decorator(func):
        if getattr(func, '__pytest_wrapped__', None):
            function = func.__pytest_wrapped__.obj
        else:
            function = func
        function.__TestLite_testcase_key__ = TestLite_id
        return func
    
    return decorator




import re

from datetime import datetime
from enum import Enum
from dataclasses import dataclass



class STATUS(str, Enum):
    SKIP = 'skip'
    PASSED = 'passed'
    FAIL = 'fail'
    ERROR = 'error'



@dataclass
class FixtureRunResult:
    result: str = None # Что возвращает фикстура
    status: STATUS = None
    error: str = None



@dataclass
class TestLiteFixtureReport:
    id: str
    nodeid: str 
    name: str = None
    cached_result: tuple = None
    before_start_time: float = None
    before_stop_time: float = None
    after_start_time: float = None
    after_stop_time: float = None
    _after_error = None


    @property
    def before_duration(self):
        if self.before_stop_time is not None:
            return self.before_stop_time - self.before_start_time
        else:
            return 0 - self.before_start_time
    

    @property
    def after_duration(self):
        if self.after_stop_time is not None and self.after_start_time is not None:
            return self.after_stop_time - self.after_start_time
        else:
            return None
        

    @property
    def before_status(self):
        if self.cached_result[2] is None:
            return FixtureRunResult(
                result=self.cached_result[0],
                status=STATUS.PASSED
            )
        else:
            return FixtureRunResult(
                status=STATUS.ERROR,
                error=self.cached_result[2]
            )
    

    @property
    def after_status(self):
        return self._after_error
    

    @after_status.setter
    def after_status(self, exc_val):
        if exc_val is not None:
            self._after_error = FixtureRunResult(
                status=STATUS.ERROR,
                error=exc_val
            )
        else:
            self._after_error = FixtureRunResult(
                status=STATUS.PASSED
            )



@dataclass
class TestLiteTestReport:
    nodeid: str
    testcase_key: str = None
    status: str = None
    startime_timestamp: float = None
    stoptime_timestamp: float = None
    report: str = None
    log: str = None
    params: str = None
    skipreason: str = None
    precondition_status: str = None
    postcondition_status: str = None
    step_number_with_error: int = None

    _fixturelist: list[TestLiteFixtureReport] = None


    @property
    def parametrize_name(self):
        name = re.search('\[.*\]', self.nodeid)
        if name is None:
            return None
        else:
            return name[0]


    @property
    def duration(self):
        if self.stoptime_timestamp is not None:
            return round(self.stoptime_timestamp - self.startime_timestamp, 2)
        else:
            return None


    @property    
    def startime_readable(self):
        if self.startime_timestamp is not None:
            return datetime.fromtimestamp(self.startime_timestamp)
        else:
            return None
        
    @property
    def stoptime_readable(self):
        if self.stoptime_timestamp is not None:
            return datetime.fromtimestamp(self.stoptime_timestamp)
        else:
            return None
        

    @property
    def fixtures(self):
        fixture_dict = {
            'before': [],
            'after': []
        }
        fixtures = self._fixturelist
        for fixture in fixtures:
            fixture_dict['before'].append(fixture)
            if fixture.after_duration is not None:
                fixture_dict['after'].append(fixture)
        return fixture_dict
    

    def add_log(self, log):
        self.log = self.log + log



import os
import time
import shutil
import pickle
import threading
import requests

from ._config import CONFIG
from ._helper import get_time
from ._models import TestLiteTestReport, TestLiteFixtureReport
from ._serializers import TestReportJSONEncoder



class fixture_after_save:
    
    def __init__(self, fixture_function, id, nodeid):
        self.fixture_function = fixture_function
        self.id = id
        self.nodeid = nodeid
        self.fixture_report = TestLiteFixtureReports.get_fixture_report(id, nodeid)


    def __call__(self, *args, **kwargs):
        with self:
            return self.fixture_function(*args, **kwargs)
        

    def __enter__(self):
        self.fixture_report.after_start_time = get_time()
        TestLiteFixtureReports.save_fixture_report(self.id, self.fixture_report)


    def __exit__(self, exc_type, exc_val, exc_tb):
        self.fixture_report.after_stop_time = get_time()
        self.fixture_report.after_status = exc_val
        TestLiteFixtureReports.save_fixture_report(self.id, self.fixture_report)



class SingletonMetaClass(type):
    _instances = {}
    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]
    


class TestLiteFixtureReports:

    __metaclass__ = SingletonMetaClass
    FixtureReports:dict[str, TestLiteFixtureReport] = {}


    @property
    def thr_context(self) -> dict[str, TestLiteTestReport]|dict[None]:
        if self._thr == threading.current_thread():
            return self.FixtureReports
        else:
            return {}
        

    def __init__(self):
        self._thr = threading.current_thread()


    @classmethod
    def get_all_fixtures_by_nodeid(cls, nodeid):
        return [item for item in cls.FixtureReports.values() if item.nodeid == nodeid]
        

    @classmethod    
    def get_fixture_report(cls, id: str, nodeid: str):
        '''
        id - уникальный идентификатор именно этой фикстуры для этого теста
        '''
        fixture_report = cls.FixtureReports.get(id)
        if fixture_report is None:
            return TestLiteFixtureReport(id=id, nodeid=nodeid)
        return fixture_report
    

    @classmethod
    def save_fixture_report(cls, id,  FixtureReport: TestLiteFixtureReport):
        '''
        id - уникальный идентификатор именно этой фикстуры для этого теста
        '''
        cls.FixtureReports.update({
            id: FixtureReport
        })
 


class TestLiteTestReports:

    __metaclass__ = SingletonMetaClass

    TestReports:dict[str, TestLiteTestReport] = {}
    save_pickle_file = 'TestLiteTemp'


    @property
    def thr_context(self) -> dict[str, TestLiteTestReport]|dict[None]:
        if self._thr == threading.current_thread():
            return self.TestReports
        else:
            return {}

    def __init__(self):
        self._thr = threading.current_thread()
        

    @classmethod    
    def get_test_report(cls, nodeid: str):
        test_report = cls.TestReports.get(nodeid)
        if test_report is None:
            return TestLiteTestReport(nodeid)
        return test_report
    
    @classmethod
    def save_test_report(cls, TestReport: TestLiteTestReport):
        test_report = TestReport
        test_report._fixturelist = TestLiteFixtureReports.get_all_fixtures_by_nodeid(TestReport.nodeid)
        cls.TestReports.update({
            TestReport.nodeid: test_report
        })
        


class TestLiteFinalReport:

    def __init__(self, report):
        self.report = report
        self.json_report = None

    def __call__(self) -> list[TestLiteTestReport]:
        return self.report
    
    def __repr__(self):
        return str(self.report)
    
    def __iter__(self):
        yield self.report

    @property
    def json(self):
        if self.json_report is None:
            self.json_report = TestReportJSONEncoder().encode(self.report)
        return self.json_report
    
    def save_json_file(self, file_name):
        with open(file_name, 'w') as file:
            file.write(self.json)

    def send_json_in_TestLite(self, testsuite):
        response = requests.post(
            url=f'{CONFIG().TESTLITEURL}/api/v1/project/{testsuite.split("-")[0]}/testsuite/{testsuite}/save',
            data=self.json,
            headers={
                'Content-Type': 'application/json'
            }
        )
      

class TestLiteReportManager:

    def __init__(self):
        self.reports = TestLiteTestReports().thr_context
        if not os.path.exists(CONFIG().REPORTSDIRNAME):
            os.mkdir(CONFIG().REPORTSDIRNAME)


    def save_report(self):
        match CONFIG().REPORTSSAVETYPE.upper():
            case 'TXT':
                self._save_report_as_txt_file()
            case 'BINARY':
                self._save_report_as_binary_file()


    def get_reports(self) -> TestLiteFinalReport:
        report = None
        match CONFIG().REPORTSSAVETYPE.upper():
            case 'TXT':
                report = self._read_reports_from_txt_files()
            case 'BINARY':
                report = self._read_reports_from_binary_files()
        
        if CONFIG().DELETEREPORTSDIR:
            shutil.rmtree(CONFIG().REPORTSDIRNAME)

        return TestLiteFinalReport(report)


    def _save_report_as_txt_file(self):
        with open(f'{CONFIG().REPORTSDIRNAME}/{str(threading.current_thread()).replace('<','').replace('>','')}.txt', 'w') as file:
            file.write(str(self.reports))

    
    def _save_report_as_binary_file(self):
        with open(f'{CONFIG().REPORTSDIRNAME}/{str(threading.current_thread()).replace('<','').replace('>','')}.data', 'wb') as file:
            file.write(pickle.dumps(self.reports))
    

    def _read_reports_from_binary_files(self):
        final_report = []
        listdir = os.listdir(CONFIG().REPORTSDIRNAME)
        for report_file_name in listdir:
            with open(f'{CONFIG().REPORTSDIRNAME}/{report_file_name}', 'rb') as file:
                final_report += [value for key, value in pickle.load(file).items()]
        return final_report
    
    
    def _read_reports_from_txt_files(self):
        final_report = []
        listdir = os.listdir(CONFIG().REPORTSDIRNAME)
        for report_file_name in listdir:
            with open(f'{CONFIG().REPORTSDIRNAME}/{report_file_name}', 'rb') as file:
                final_report += [value for key, value in dict(file.read()).items()]
        return final_report


from setuptools import setup, find_packages

setup(
    name='TestLitePytest',  # Replace with your package’s name
    version='0.2.1',
    package_dir={"": "src"},
    packages=['testlite_pytest', 'testlite'],
    install_requires=[
        'requests>=2.32.3',
        'pytest>=8.3.2'
    ],
    author='Dmitry Skryabin',  
    author_email='skryabind98@gmail.com',
    description='Pytest adaptor for TestLite TMS system',
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',  # License type
        'Operating System :: OS Independent',
        "Framework :: Pytest"
    ],
    python_requires='>=3.10',
)


import os
import configparser


class SingletonMetaClass(type):
    _instances = {}
    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]


class CONFIG:
    __metaclass__ = SingletonMetaClass

    _REPORTSDIRNAME = 'TestLiteReports'
    _DELETEREPORTSDIR = True
    _REPORTSSAVETYPE = 'BINARY'
    _TESTLITEURL = 'http://127.0.0.1:8000'

    def __init__(self):
        self.config = configparser.ConfigParser()
        if os.path.exists('TestLiteConfig.ini'):
            self.have_config_file = True
            self.config.read('TestLiteConfig.ini')
        else:
            self.have_config_file = False
 
    @property
    def TESTLITEURL(self):
        config_value = self.config.get('TestLiteConfig', 'TESTLITEURL', fallback=None)
        if config_value is not None:
            return config_value
        else:
            return self._TESTLITEURL
        
    @property
    def DELETEREPORTSDIR(self):
        config_value = self.config.get('TestLiteConfig', 'DELETEREPORTSDIR', fallback=None)
        if config_value is not None:
            return config_value
        else:
            return self._DELETEREPORTSDIR
        
    @property
    def REPORTSDIRNAME(self):
        config_value = self.config.get('TestLiteConfig', 'REPORTSDIRNAME', fallback=None)
        if config_value is not None:
            return config_value
        else:
            return self._REPORTSDIRNAME
        
    @property
    def REPORTSSAVETYPE(self):
        config_value = self.config.get('TestLiteConfig', 'REPORTSSAVETYPE', fallback=None)
        if config_value is not None:
            return config_value
        else:
            return self._REPORTSSAVETYPE



    
def TestLite_testcase_key(item):
    return getattr(
        getattr(item, "obj", None),
        "__TestLite_testcase_key__",
        None
    )


def get_step_number_with_error(longreprtext: str) -> int|None:
    step_number = 0
    for i, line in enumerate(longreprtext.split('\n')):
        if '#TestLiteStep' in line:
            step_number += 1
    return step_number if step_number !=0 else None

